import gleam/dynamic/decode.{type Decoder}
import models/faction.{type Faction}
import models/faction_symbol.{type FactionSymbol}
import utils/auth.{type AgentToken, AgentAuth}
import utils/request as request_utils
import utils/response as response_utils
import utils/types.{type ApiResponse, type PagedData}

pub type ListFactionsResponse {
  ListFactionsResponse(factions: List(Faction))
}

fn list_factions_response_decoder() -> Decoder(ListFactionsResponse) {
  use factions <- decode.then(decode.list(faction.decoder()))
  decode.success(ListFactionsResponse(factions:))
}

pub fn list_factions(
  token: AgentToken,
  page: Int,
  limit: Int,
) -> ApiResponse(PagedData(ListFactionsResponse)) {
  let request =
    request_utils.get_page(AgentAuth(token), "/factions", page, limit)
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.paged_data_parser(
        response.body,
        list_factions_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type GetFactionResponse {
  GetFactionResponse(faction: Faction)
}

fn get_faction_response_decoder() -> Decoder(GetFactionResponse) {
  use faction <- decode.then(faction.decoder())
  decode.success(GetFactionResponse(faction:))
}

pub fn get_faction(
  token: AgentToken,
  symbol: FactionSymbol,
) -> ApiResponse(GetFactionResponse) {
  let request =
    request_utils.get(
      AgentAuth(token),
      "/factions/" <> faction_symbol.to_string(symbol),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(response.body, get_faction_response_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type FactionReputation {
  FactionReputation(symbol: FactionSymbol, reputation: Int)
}

fn faction_reputation_decoder() {
  use symbol <- decode.field("symbol", faction_symbol.decoder())
  use reputation <- decode.field("reputation", decode.int)
  decode.success(FactionReputation(symbol:, reputation:))
}

pub type GetMyFactionsResponse {
  GetMyFactionsResponse(faction_reputations: List(FactionReputation))
}

fn get_my_factions_response_decoder() -> Decoder(GetMyFactionsResponse) {
  use faction_reputations <- decode.then(
    decode.list(faction_reputation_decoder()),
  )
  decode.success(GetMyFactionsResponse(faction_reputations:))
}

pub fn get_my_factions(
  token: AgentToken,
  page: Int,
  limit: Int,
) -> ApiResponse(PagedData(GetMyFactionsResponse)) {
  let request =
    request_utils.get_page(AgentAuth(token), "/my/factions", page, limit)
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.paged_data_parser(
        response.body,
        get_my_factions_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}
