import gleam/dynamic/decode.{type Decoder}
import gleam/json
import gleam/option.{type Option}
import models/agent.{type Agent}
import models/chart.{type Chart}
import models/chart_transaction.{type ChartTransaction}
import models/contract.{type Contract}
import models/cooldown.{type Cooldown}
import models/extraction.{type Extraction}
import models/market_transaction.{type MarketTransaction}
import models/module_symbol.{type ModuleSymbol}
import models/mount_symbol.{type MountSymbol}
import models/refinement_produce.{type RefinementProduce}
import models/refinement_yield.{type RefinementYield}
import models/repair_transaction.{type RepairTransaction}
import models/scanned_ship.{type ScannedShip}
import models/scanned_system.{type ScannedSystem}
import models/scanned_waypoint.{type ScannedWaypoint}
import models/scrap_transaction.{type ScrapTransaction}
import models/ship.{type Ship}
import models/ship_cargo.{type ShipCargo}
import models/ship_condition_event.{type ShipConditionEvent}
import models/ship_fuel.{type ShipFuel}
import models/ship_modification_transaction.{type ShipModificationTransaction}
import models/ship_module.{type ShipModule}
import models/ship_mount.{type ShipMount}
import models/ship_nav.{type ShipNav}
import models/ship_nav_flight_mode.{type ShipNavFlightMode}
import models/ship_symbol.{type ShipSymbol}
import models/ship_type.{type ShipType}
import models/shipyard_transaction.{type ShipyardTransaction}
import models/siphon.{type Siphon}
import models/survey.{type Survey}
import models/trade_symbol.{type TradeSymbol}
import models/waypoint.{type Waypoint}
import models/waypoint_modifier.{type WaypointModifier}
import models/waypoint_symbol.{type WaypointSymbol}
import utils/auth.{type AgentToken, AgentAuth}
import utils/request as request_utils
import utils/response as response_utils
import utils/types.{type ApiResponse, type PagedData}

pub type ListShipsResponse {
  ListShipsResponse(ships: List(Ship))
}

pub fn list_ships_decoder() -> Decoder(ListShipsResponse) {
  use ships <- decode.then(decode.list(ship.decoder()))
  decode.success(ListShipsResponse(ships:))
}

pub fn list_ships(
  token: AgentToken,
  page: Int,
  limit: Int,
) -> ApiResponse(PagedData(ListShipsResponse)) {
  let request =
    request_utils.get_page(AgentAuth(token), "/my/ships", page, limit)
  use response <- request_utils.try_send(request)
  case response.status {
    200 -> response_utils.paged_data_parser(response.body, list_ships_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type PurchaseShipResponse {
  PurchaseShipResponse(
    ship: Ship,
    agent: Agent,
    transaction: ShipyardTransaction,
  )
}

pub fn purchase_ship_decoder() -> Decoder(PurchaseShipResponse) {
  use ship <- decode.field("ship", ship.decoder())
  use agent <- decode.field("agent", agent.decoder())
  use transaction <- decode.field("transaction", shipyard_transaction.decoder())
  decode.success(PurchaseShipResponse(ship:, agent:, transaction:))
}

pub fn purchase_ship(
  token: AgentToken,
  ship_type: ShipType,
  waypoint_symbol: WaypointSymbol,
) -> ApiResponse(PurchaseShipResponse) {
  let request =
    request_utils.post_json(
      AgentAuth(token),
      "/my/ships",
      json.object([
        #("shipType", json.string(ship_type.to_string(ship_type))),
        #("waypointSymbol", json.string(waypoint_symbol)),
      ]),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 -> response_utils.data_parser(response.body, purchase_ship_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type GetShipResponse {
  GetShipResponse(ship: Ship)
}

pub fn get_ship_decoder() -> Decoder(GetShipResponse) {
  use ship <- decode.then(ship.decoder())
  decode.success(GetShipResponse(ship:))
}

pub fn get_ship(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(GetShipResponse) {
  let request = request_utils.get(AgentAuth(token), "/my/ships/" <> ship_symbol)
  use response <- request_utils.try_send(request)
  case response.status {
    200 -> response_utils.data_parser(response.body, get_ship_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type CreateChartResponse {
  CreateChartResponse(
    chart: Chart,
    waypoint: Waypoint,
    transaction: ChartTransaction,
    agent: Agent,
  )
}

pub fn create_chart_response_decoder() -> Decoder(CreateChartResponse) {
  use chart <- decode.field("chart", chart.decoder())
  use waypoint <- decode.field("waypoint", waypoint.decoder())
  use transaction <- decode.field("transaction", chart_transaction.decoder())
  use agent <- decode.field("agent", agent.decoder())
  decode.success(CreateChartResponse(chart:, waypoint:, transaction:, agent:))
}

pub fn create_chart(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(CreateChartResponse) {
  let request =
    request_utils.post(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/chart",
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(response.body, create_chart_response_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type NegotiateContractResponse {
  NegotiateContractResponse(contract: Contract)
}

fn negotiate_contract_response_decoder() -> Decoder(NegotiateContractResponse) {
  use contract <- decode.field("contract", contract.decoder())
  decode.success(NegotiateContractResponse(contract:))
}

pub fn negotiate_contract(
  token: AgentToken,
  ship_symbol: String,
) -> ApiResponse(NegotiateContractResponse) {
  let request =
    request_utils.post(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/negotiate/contract",
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(
        response.body,
        negotiate_contract_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type GetShipCooldownResponse {
  GetShipCooldownResponse(cooldown: Option(Cooldown))
}

fn get_ship_cooldown_response_decoder() -> Decoder(GetShipCooldownResponse) {
  use cooldown <- decode.field("cooldown", cooldown.decoder())
  decode.success(GetShipCooldownResponse(cooldown: option.Some(cooldown)))
}

pub fn get_ship_cooldown(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(GetShipCooldownResponse) {
  let request =
    request_utils.get(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/cooldown",
    )
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(
        response.body,
        get_ship_cooldown_response_decoder(),
      )
    204 -> Ok(GetShipCooldownResponse(option.None))
    _ -> response_utils.error_parser(response.body)
  }
}

pub type DockShipResponse {
  DockShipResponse(nav: ShipNav)
}

fn dock_ship_response_decoder() -> Decoder(DockShipResponse) {
  use nav <- decode.field("nav", ship_nav.decoder())
  decode.success(DockShipResponse(nav:))
}

pub fn dock_ship(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(DockShipResponse) {
  let request =
    request_utils.post(AgentAuth(token), "/my/ships/" <> ship_symbol <> "/dock")
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(response.body, dock_ship_response_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type ExtractResourcesResponse {
  ExtractResourcesResponse(
    extraction: Extraction,
    cooldown: Cooldown,
    cargo: ShipCargo,
    modifiers: List(WaypointModifier),
    events: List(ShipConditionEvent),
  )
}

fn extract_resources_response_decoder() -> Decoder(ExtractResourcesResponse) {
  use extraction <- decode.field("extraction", extraction.decoder())
  use cooldown <- decode.field("cooldown", cooldown.decoder())
  use cargo <- decode.field("cargo", ship_cargo.decoder())
  use modifiers <- decode.field(
    "modifiers",
    decode.list(waypoint_modifier.decoder()),
  )
  use events <- decode.field(
    "events",
    decode.list(ship_condition_event.decoder()),
  )
  decode.success(ExtractResourcesResponse(
    extraction:,
    cooldown:,
    cargo:,
    modifiers:,
    events:,
  ))
}

pub fn extract_resources(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(ExtractResourcesResponse) {
  let request =
    request_utils.post(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/extract",
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(
        response.body,
        extract_resources_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type ExtractResourcesWithSurveyResponse {
  ExtractResourcesWithSurveyResponse(
    extraction: Extraction,
    cooldown: Cooldown,
    cargo: ShipCargo,
    modifiers: List(WaypointModifier),
    events: List(ShipConditionEvent),
  )
}

fn extract_resources_with_survey_response_decoder() -> Decoder(
  ExtractResourcesWithSurveyResponse,
) {
  use extraction <- decode.field("extraction", extraction.decoder())
  use cooldown <- decode.field("cooldown", cooldown.decoder())
  use cargo <- decode.field("cargo", ship_cargo.decoder())
  use modifiers <- decode.field(
    "modifiers",
    decode.list(waypoint_modifier.decoder()),
  )
  use events <- decode.field(
    "events",
    decode.list(ship_condition_event.decoder()),
  )
  decode.success(ExtractResourcesWithSurveyResponse(
    extraction:,
    cooldown:,
    cargo:,
    modifiers:,
    events:,
  ))
}

pub fn extract_resources_with_survey(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  survey: Survey,
) -> ApiResponse(ExtractResourcesWithSurveyResponse) {
  let request =
    request_utils.post_json(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/extract/survey",
      survey.encode(survey),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(
        response.body,
        extract_resources_with_survey_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type JettisonCargoResponse {
  JettisonCargoResponse(cargo: ShipCargo)
}

fn jettison_cargo_response_decoder() -> Decoder(JettisonCargoResponse) {
  use cargo <- decode.field("cargo", ship_cargo.decoder())
  decode.success(JettisonCargoResponse(cargo:))
}

pub fn jettison_cargo(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  trade_symbol: TradeSymbol,
  units: Int,
) -> ApiResponse(JettisonCargoResponse) {
  let request =
    request_utils.post_json(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/jettison",
      json.object([
        #("symbol", json.string(trade_symbol.to_string(trade_symbol))),
        #("units", json.int(units)),
      ]),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(
        response.body,
        jettison_cargo_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type JumpShipResponse {
  JumpShipResponse(
    nav: ShipNav,
    cooldown: Cooldown,
    transaction: MarketTransaction,
    agent: Agent,
  )
}

fn jump_ship_response_decoder() -> Decoder(JumpShipResponse) {
  use nav <- decode.field("nav", ship_nav.decoder())
  use cooldown <- decode.field("cooldown", cooldown.decoder())
  use transaction <- decode.field("transaction", market_transaction.decoder())
  use agent <- decode.field("agent", agent.decoder())
  decode.success(JumpShipResponse(nav:, cooldown:, transaction:, agent:))
}

pub fn jump_ship(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  waypoint_symbol: WaypointSymbol,
) -> ApiResponse(JumpShipResponse) {
  let request =
    request_utils.post_json(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/jump",
      json.object([#("waypointSymbol", json.string(waypoint_symbol))]),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(response.body, jump_ship_response_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type SystemScanResponse {
  SystemScanResponse(cooldown: Cooldown, systems: List(ScannedSystem))
}

fn system_scan_response_decoder() -> Decoder(SystemScanResponse) {
  use cooldown <- decode.field("cooldown", cooldown.decoder())
  use systems <- decode.field("systems", decode.list(scanned_system.decoder()))
  decode.success(SystemScanResponse(cooldown:, systems:))
}

pub fn system_scan(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(SystemScanResponse) {
  let request =
    request_utils.post(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/scan/systems",
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(response.body, system_scan_response_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type WaypointScanResponse {
  WaypointScanResponse(cooldown: Cooldown, waypoints: List(ScannedWaypoint))
}

fn waypoint_scan_response_decoder() -> Decoder(WaypointScanResponse) {
  use cooldown <- decode.field("cooldown", cooldown.decoder())
  use waypoints <- decode.field(
    "waypoints",
    decode.list(scanned_waypoint.decoder()),
  )
  decode.success(WaypointScanResponse(cooldown:, waypoints:))
}

pub fn waypoint_scan(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(WaypointScanResponse) {
  let request =
    request_utils.post(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/scan/waypoints",
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(
        response.body,
        waypoint_scan_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type ShipScanResponse {
  ShipScanResponse(cooldown: Cooldown, ships: List(ScannedShip))
}

fn ship_scan_response_decoder() -> Decoder(ShipScanResponse) {
  use cooldown <- decode.field("cooldown", cooldown.decoder())
  use ships <- decode.field("ships", decode.list(scanned_ship.decoder()))
  decode.success(ShipScanResponse(cooldown:, ships:))
}

pub fn scan_ships(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(ShipScanResponse) {
  let request =
    request_utils.post(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/scan/ships",
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(response.body, ship_scan_response_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type ScrapShipResponse {
  ScrapShipResponse(agent: Agent, transaction: ScrapTransaction)
}

fn scrap_ship_response_decoder() -> Decoder(ScrapShipResponse) {
  use agent <- decode.field("agent", agent.decoder())
  use transaction <- decode.field("transaction", scrap_transaction.decoder())
  decode.success(ScrapShipResponse(agent:, transaction:))
}

pub fn scrap_ship(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(ScrapShipResponse) {
  let request =
    request_utils.post(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/scrap",
    )
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(response.body, scrap_ship_response_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type GetScrapShipResponse {
  GetScrapShipResponse(transaction: ScrapTransaction)
}

fn get_scrap_ship_response_decoder() -> Decoder(GetScrapShipResponse) {
  use transaction <- decode.field("transaction", scrap_transaction.decoder())
  decode.success(GetScrapShipResponse(transaction:))
}

pub fn get_scrap_ship(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(GetScrapShipResponse) {
  let request =
    request_utils.get(AgentAuth(token), "/my/ships/" <> ship_symbol <> "/scrap")
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(
        response.body,
        get_scrap_ship_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type NavigateShipResponse {
  NavigateShipResponse(
    nav: ShipNav,
    fuel: ShipFuel,
    events: List(ShipConditionEvent),
  )
}

fn navigate_ship_response_decoder() -> Decoder(NavigateShipResponse) {
  use nav <- decode.field("nav", ship_nav.decoder())
  use fuel <- decode.field("fuel", ship_fuel.decoder())
  use events <- decode.field(
    "events",
    decode.list(ship_condition_event.decoder()),
  )
  decode.success(NavigateShipResponse(nav:, fuel:, events:))
}

pub fn navigate_ship(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  waypoint_symbol: WaypointSymbol,
) -> ApiResponse(NavigateShipResponse) {
  let request =
    request_utils.post_json(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/navigate",
      json.object([#("waypointSymbol", json.string(waypoint_symbol))]),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(
        response.body,
        navigate_ship_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type WarpShipResponse {
  WarpShipResponse(
    nav: ShipNav,
    fuel: ShipFuel,
    events: List(ShipConditionEvent),
  )
}

fn warp_ship_response_decoder() -> Decoder(WarpShipResponse) {
  use nav <- decode.field("nav", ship_nav.decoder())
  use fuel <- decode.field("fuel", ship_fuel.decoder())
  use events <- decode.field(
    "events",
    decode.list(ship_condition_event.decoder()),
  )
  decode.success(WarpShipResponse(nav:, fuel:, events:))
}

pub fn warp_ship(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  waypoint_symbol: WaypointSymbol,
) -> ApiResponse(WarpShipResponse) {
  let request =
    request_utils.post_json(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/warp",
      json.object([#("waypointSymbol", json.string(waypoint_symbol))]),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(response.body, warp_ship_response_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type OrbitShipResponse {
  OrbitShipResponse(nav: ShipNav)
}

fn orbit_ship_response_decoder() -> Decoder(OrbitShipResponse) {
  use nav <- decode.field("nav", ship_nav.decoder())
  decode.success(OrbitShipResponse(nav:))
}

pub fn orbit_ship(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(OrbitShipResponse) {
  let request =
    request_utils.post(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/orbit",
    )
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(response.body, orbit_ship_response_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type PurchaseCargoResponse {
  PurchaseCargoResponse(
    cargo: ShipCargo,
    transaction: MarketTransaction,
    agent: Agent,
  )
}

fn purchase_cargo_response_decoder() -> Decoder(PurchaseCargoResponse) {
  use cargo <- decode.field("cargo", ship_cargo.decoder())
  use transaction <- decode.field("transaction", market_transaction.decoder())
  use agent <- decode.field("agent", agent.decoder())
  decode.success(PurchaseCargoResponse(cargo:, transaction:, agent:))
}

pub fn purchase_cargo(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  trade_symbol: TradeSymbol,
) -> ApiResponse(PurchaseCargoResponse) {
  let request =
    request_utils.post_json(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/purchase",
      json.object([
        #("symbol", json.string(trade_symbol.to_string(trade_symbol))),
      ]),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(
        response.body,
        purchase_cargo_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type ShipRefineResponse {
  ShipRefineResponse(
    cargo: ShipCargo,
    cooldown: Cooldown,
    produced: List(RefinementYield),
    consumed: List(RefinementYield),
  )
}

fn ship_refine_response_decoder() -> Decoder(ShipRefineResponse) {
  use cargo <- decode.field("cargo", ship_cargo.decoder())
  use cooldown <- decode.field("cooldown", cooldown.decoder())
  use produced <- decode.field(
    "produced",
    decode.list(refinement_yield.decoder()),
  )
  use consumed <- decode.field(
    "consumed",
    decode.list(refinement_yield.decoder()),
  )
  decode.success(ShipRefineResponse(cargo:, cooldown:, produced:, consumed:))
}

pub fn ship_refine(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  produce: RefinementProduce,
) -> ApiResponse(ShipRefineResponse) {
  let request =
    request_utils.post_json(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/refine",
      json.object([
        #("produce", json.string(refinement_produce.to_string(produce))),
      ]),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(response.body, ship_refine_response_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type RefuelShipResponse {
  RefuelShipResponse(
    agent: Agent,
    fuel: ShipFuel,
    cargo: ShipCargo,
    transaction: MarketTransaction,
  )
}

pub fn refuel_ship_response_decoder() -> Decoder(RefuelShipResponse) {
  use agent <- decode.field("agent", agent.decoder())
  use fuel <- decode.field("fuel", ship_fuel.decoder())
  use cargo <- decode.field("cargo", ship_cargo.decoder())
  use transaction <- decode.field("transaction", market_transaction.decoder())
  decode.success(RefuelShipResponse(agent:, fuel:, cargo:, transaction:))
}

pub fn refuel_ship(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  units: Option(Int),
  from_cargo: Option(Bool),
) -> ApiResponse(RefuelShipResponse) {
  let request =
    request_utils.post_json(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/refuel",
      json.object([
        #("units", json.nullable(units, json.int)),
        #("fromCargo", json.nullable(from_cargo, json.bool)),
      ]),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(response.body, refuel_ship_response_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type RepairShipResponse {
  RepairShipResponse(agent: Agent, ship: Ship, transaction: RepairTransaction)
}

pub fn repair_ship_response_decoder() -> Decoder(RepairShipResponse) {
  use agent <- decode.field("agent", agent.decoder())
  use ship <- decode.field("ship", ship.decoder())
  use transaction <- decode.field("transaction", repair_transaction.decoder())
  decode.success(RepairShipResponse(agent:, ship:, transaction:))
}

pub fn repair_ship(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(RepairShipResponse) {
  let request =
    request_utils.post(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/repair",
    )
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(response.body, repair_ship_response_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type GetRepairShipResponse {
  GetRepairShipResponse(transaction: RepairTransaction)
}

pub fn get_repair_ship_response_decoder() -> Decoder(GetRepairShipResponse) {
  use transaction <- decode.field("transaction", repair_transaction.decoder())
  decode.success(GetRepairShipResponse(transaction:))
}

pub fn get_repair_ship(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(GetRepairShipResponse) {
  let request =
    request_utils.get(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/repair",
    )
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(
        response.body,
        get_repair_ship_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type SellCargoResponse {
  SellCargoResponse(
    cargo: ShipCargo,
    transaction: MarketTransaction,
    agent: Agent,
  )
}

pub fn sell_cargo_response_decoder() -> Decoder(SellCargoResponse) {
  use cargo <- decode.field("cargo", ship_cargo.decoder())
  use transaction <- decode.field("transaction", market_transaction.decoder())
  use agent <- decode.field("agent", agent.decoder())
  decode.success(SellCargoResponse(cargo:, transaction:, agent:))
}

pub fn sell_cargo(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  trade_symbol: TradeSymbol,
  units: Int,
) -> ApiResponse(SellCargoResponse) {
  let request =
    request_utils.post_json(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/sell",
      json.object([
        #("symbol", json.string(trade_symbol.to_string(trade_symbol))),
        #("units", json.int(units)),
      ]),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(response.body, sell_cargo_response_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type SiphonResourcesResponse {
  SiphonResourcesResponse(
    siphon: Siphon,
    cooldown: Cooldown,
    cargo: ShipCargo,
    events: List(ShipConditionEvent),
  )
}

pub fn siphon_resources_response_decoder() -> Decoder(SiphonResourcesResponse) {
  use siphon <- decode.field("siphon", siphon.decoder())
  use cooldown <- decode.field("cooldown", cooldown.decoder())
  use cargo <- decode.field("cargo", ship_cargo.decoder())
  use events <- decode.field(
    "events",
    decode.list(ship_condition_event.decoder()),
  )
  decode.success(SiphonResourcesResponse(siphon:, cooldown:, cargo:, events:))
}

pub fn siphon_resources(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(SiphonResourcesResponse) {
  let request =
    request_utils.post(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/siphon",
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(
        response.body,
        siphon_resources_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type CreateSurveyResponse {
  CreateSurveyResponse(cooldown: Cooldown, surveys: List(Survey))
}

pub fn create_survey_response_decoder() -> Decoder(CreateSurveyResponse) {
  use cooldown <- decode.field("cooldown", cooldown.decoder())
  use surveys <- decode.field("surveys", decode.list(survey.decoder()))
  decode.success(CreateSurveyResponse(cooldown:, surveys:))
}

pub fn create_survey(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(CreateSurveyResponse) {
  let request =
    request_utils.post(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/survey",
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(
        response.body,
        create_survey_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type TransferCargoResponse {
  TransferCargoResponse(cargo: ShipCargo, target_cargo: ShipCargo)
}

pub fn transfer_cargo_response_decoder() -> Decoder(TransferCargoResponse) {
  use cargo <- decode.field("cargo", ship_cargo.decoder())
  use target_cargo <- decode.field("targetCargo", ship_cargo.decoder())
  decode.success(TransferCargoResponse(cargo:, target_cargo:))
}

pub fn transfer_cargo(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  trade_symbol: TradeSymbol,
  units: Int,
  target_ship_symbol: ShipSymbol,
) -> ApiResponse(TransferCargoResponse) {
  let request =
    request_utils.post_json(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/transfer",
      json.object([
        #("tradeSymbol", json.string(trade_symbol.to_string(trade_symbol))),
        #("units", json.int(units)),
        #("shipSymbol", json.string(target_ship_symbol)),
      ]),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(
        response.body,
        transfer_cargo_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type GetShipCargoResponse {
  GetShipCargoResponse(cargo: ShipCargo)
}

pub fn get_ship_cargo_response_decoder() -> Decoder(GetShipCargoResponse) {
  use cargo <- decode.then(ship_cargo.decoder())
  decode.success(GetShipCargoResponse(cargo:))
}

pub fn get_my_ship_cargo(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(GetShipCargoResponse) {
  let request =
    request_utils.get(AgentAuth(token), "/my/ships/" <> ship_symbol <> "/cargo")
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(
        response.body,
        get_ship_cargo_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type GetShipModulesResponse {
  GetShipModulesResponse(modules: List(ShipModule))
}

pub fn get_ship_modules_response_decoder() -> Decoder(GetShipModulesResponse) {
  use modules <- decode.then(decode.list(ship_module.decoder()))
  decode.success(GetShipModulesResponse(modules:))
}

pub fn get_ship_modules(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(GetShipModulesResponse) {
  let request =
    request_utils.get(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/modules",
    )
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(
        response.body,
        get_ship_modules_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type InstallShipModuleResponse {
  InstallShipModuleResponse(
    agent: Agent,
    modules: List(ShipModule),
    cargo: ShipCargo,
    transaction: ShipModificationTransaction,
  )
}

pub fn install_ship_module_response_decoder() -> Decoder(
  InstallShipModuleResponse,
) {
  use agent <- decode.field("agent", agent.decoder())
  use modules <- decode.field("modules", decode.list(ship_module.decoder()))
  use cargo <- decode.field("cargo", ship_cargo.decoder())
  use transaction <- decode.field(
    "transaction",
    ship_modification_transaction.decoder(),
  )
  decode.success(InstallShipModuleResponse(
    agent:,
    modules:,
    cargo:,
    transaction:,
  ))
}

pub fn install_ship_module(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  module_symbol: ModuleSymbol,
) -> ApiResponse(InstallShipModuleResponse) {
  let request =
    request_utils.post_json(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/modules/install",
      json.object([
        #("symbol", json.string(module_symbol.to_string(module_symbol))),
      ]),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(
        response.body,
        install_ship_module_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type RemoveShipModuleResponse {
  RemoveShipModuleResponse(
    agent: Agent,
    modules: List(ShipModule),
    cargo: ShipCargo,
    transaction: ShipModificationTransaction,
  )
}

pub fn remove_ship_module_response_decoder() -> Decoder(
  RemoveShipModuleResponse,
) {
  use agent <- decode.field("agent", agent.decoder())
  use modules <- decode.field("modules", decode.list(ship_module.decoder()))
  use cargo <- decode.field("cargo", ship_cargo.decoder())
  use transaction <- decode.field(
    "transaction",
    ship_modification_transaction.decoder(),
  )
  decode.success(RemoveShipModuleResponse(
    agent:,
    modules:,
    cargo:,
    transaction:,
  ))
}

pub fn remove_ship_module(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  module_symbol: ModuleSymbol,
) -> ApiResponse(RemoveShipModuleResponse) {
  let request =
    request_utils.post_json(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/modules/remove",
      json.object([
        #("symbol", json.string(module_symbol.to_string(module_symbol))),
      ]),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(
        response.body,
        remove_ship_module_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type GetShipMountsResponse {
  GetShipMountsResponse(mounts: List(ShipMount))
}

pub fn get_ship_mounts_response_decoder() -> Decoder(GetShipMountsResponse) {
  use mounts <- decode.then(decode.list(ship_mount.decoder()))
  decode.success(GetShipMountsResponse(mounts:))
}

pub fn get_ship_mounts(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(GetShipMountsResponse) {
  let request =
    request_utils.get(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/mounts",
    )
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(
        response.body,
        get_ship_mounts_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type InstallShipMountResponse {
  InstallShipMountResponse(
    agent: Agent,
    mounts: List(ShipMount),
    cargo: ShipCargo,
    transaction: ShipModificationTransaction,
  )
}

pub fn install_ship_mount_response_decoder() -> Decoder(
  InstallShipMountResponse,
) {
  use agent <- decode.field("agent", agent.decoder())
  use mounts <- decode.field("mounts", decode.list(ship_mount.decoder()))
  use cargo <- decode.field("cargo", ship_cargo.decoder())
  use transaction <- decode.field(
    "transaction",
    ship_modification_transaction.decoder(),
  )
  decode.success(InstallShipMountResponse(agent:, mounts:, cargo:, transaction:))
}

pub fn install_ship_mount(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  mount_symbol: MountSymbol,
) -> ApiResponse(InstallShipMountResponse) {
  let request =
    request_utils.post_json(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/mounts/install",
      json.object([
        #("symbol", json.string(mount_symbol.to_string(mount_symbol))),
      ]),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(
        response.body,
        install_ship_mount_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type RemoveShipMountResponse {
  RemoveShipMountResponse(
    agent: Agent,
    mounts: List(ShipMount),
    cargo: ShipCargo,
    transaction: ShipModificationTransaction,
  )
}

pub fn remove_ship_mount_response_decoder() -> Decoder(RemoveShipMountResponse) {
  use agent <- decode.field("agent", agent.decoder())
  use mounts <- decode.field("mounts", decode.list(ship_mount.decoder()))
  use cargo <- decode.field("cargo", ship_cargo.decoder())
  use transaction <- decode.field(
    "transaction",
    ship_modification_transaction.decoder(),
  )
  decode.success(RemoveShipMountResponse(agent:, mounts:, cargo:, transaction:))
}

pub fn remove_ship_mount(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  mount_symbol: MountSymbol,
) -> ApiResponse(RemoveShipMountResponse) {
  let request =
    request_utils.post_json(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/mounts/remove",
      json.object([
        #("symbol", json.string(mount_symbol.to_string(mount_symbol))),
      ]),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    201 ->
      response_utils.data_parser(
        response.body,
        remove_ship_mount_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}

pub type GetShipNavResponse {
  GetShipNavResponse(nav: ShipNav)
}

pub fn get_ship_nav_response_decoder() -> Decoder(GetShipNavResponse) {
  use nav <- decode.then(ship_nav.decoder())
  decode.success(GetShipNavResponse(nav:))
}

pub fn get_ship_nav(
  token: AgentToken,
  ship_symbol: ShipSymbol,
) -> ApiResponse(GetShipNavResponse) {
  let request =
    request_utils.get(AgentAuth(token), "/my/ships/" <> ship_symbol <> "/nav")
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(response.body, get_ship_nav_response_decoder())
    _ -> response_utils.error_parser(response.body)
  }
}

pub type PatchShipNavResponse {
  PatchShipNavResponse(
    nav: ShipNav,
    fuel: ShipFuel,
    events: List(ShipConditionEvent),
  )
}

pub fn patch_ship_nav_response_decoder() -> Decoder(PatchShipNavResponse) {
  use nav <- decode.field("nav", ship_nav.decoder())
  use fuel <- decode.field("fuel", ship_fuel.decoder())
  use events <- decode.field(
    "events",
    decode.list(ship_condition_event.decoder()),
  )
  decode.success(PatchShipNavResponse(nav:, fuel:, events:))
}

pub fn patch_ship_nav(
  token: AgentToken,
  ship_symbol: ShipSymbol,
  flight_mode: ShipNavFlightMode,
) -> ApiResponse(PatchShipNavResponse) {
  let request =
    request_utils.patch_json(
      AgentAuth(token),
      "/my/ships/" <> ship_symbol <> "/nav",
      json.object([
        #(
          "flightMode",
          json.string(ship_nav_flight_mode.to_string(flight_mode)),
        ),
      ]),
    )
  use response <- request_utils.try_send(request)
  case response.status {
    200 ->
      response_utils.data_parser(
        response.body,
        patch_ship_nav_response_decoder(),
      )
    _ -> response_utils.error_parser(response.body)
  }
}
