import gleam/bit_array
import gleam/http.{type Method}
import gleam/http/request.{type Request}
import gleam/http/response.{type Response}
import gleam/httpc
import gleam/int
import gleam/json.{type Json}
import gleam/option
import gleam/result
import gleam/string_tree
import utils/auth.{type AuthMethod}
import utils/types.{type ApiResponse}

const base_request = request.Request(
  method: http.Get,
  headers: [],
  body: <<>>,
  scheme: http.Https,
  host: "api.spacetraders.io",
  port: option.None,
  path: "/v2",
  query: option.None,
)

fn set_body(
  req: request.Request(a),
  content_type: String,
  body: BitArray,
) -> request.Request(BitArray) {
  req
  |> request.set_header("Content-Type", content_type)
  |> request.set_body(body)
}

fn set_text_body(req: Request(a), text: String) {
  req |> set_body("text/plain", bit_array.from_string(text))
}

fn set_json_body(req: Request(a), json: Json) {
  req
  |> set_body(
    "application/json",
    json
      |> json.to_string_tree
      |> string_tree.to_string
      |> bit_array.from_string,
  )
}

fn make_request(method: Method, auth_method: AuthMethod, path: String) {
  base_request
  |> request.set_method(method)
  |> auth.set_auth(auth_method)
  |> request.set_path(base_request.path <> path)
}

pub fn get(auth_method: AuthMethod, path: String) {
  make_request(http.Get, auth_method, path)
}

pub fn get_query(
  auth_method: AuthMethod,
  path: String,
  query: List(#(String, String)),
) {
  make_request(http.Get, auth_method, path) |> request.set_query(query)
}

pub fn get_page(auth_method: AuthMethod, path: String, page: Int, limit: Int) {
  make_request(http.Get, auth_method, path)
  |> request.set_query([
    #("page", int.to_string(page)),
    #("limit", int.to_string(limit)),
  ])
}

pub fn get_page_query(
  auth_method: AuthMethod,
  path: String,
  page: Int,
  limit: Int,
  query: List(#(String, String)),
) {
  make_request(http.Get, auth_method, path)
  |> request.set_query([
    #("page", int.to_string(page)),
    #("limit", int.to_string(limit)),
    ..query
  ])
}

pub fn post(auth_method: AuthMethod, path: String) {
  make_request(http.Post, auth_method, path)
}

pub fn post_text(auth_method: AuthMethod, path: String, body: String) {
  make_request(http.Post, auth_method, path) |> set_text_body(body)
}

pub fn post_json(auth_method: AuthMethod, path: String, body: Json) {
  make_request(http.Post, auth_method, path) |> set_json_body(body)
}

pub fn patch(auth_method: AuthMethod, path: String) {
  make_request(http.Patch, auth_method, path)
}

pub fn patch_text(auth_method: AuthMethod, path: String, body: String) {
  make_request(http.Patch, auth_method, path) |> set_text_body(body)
}

pub fn patch_json(auth_method: AuthMethod, path: String, body: Json) {
  make_request(http.Patch, auth_method, path) |> set_json_body(body)
}

pub fn put(auth_method: AuthMethod, path: String) {
  make_request(http.Put, auth_method, path)
}

pub fn put_text(auth_method: AuthMethod, path: String, body: String) {
  make_request(http.Put, auth_method, path) |> set_text_body(body)
}

pub fn put_json(auth_method: AuthMethod, path: String, body: Json) {
  make_request(http.Put, auth_method, path) |> set_json_body(body)
}

pub fn delete(auth_method: AuthMethod, path: String) {
  make_request(http.Delete, auth_method, path)
}

pub fn send(request: Request(BitArray)) {
  case httpc.send_bits(request) {
    Ok(res) -> Ok(res)
    Error(err) -> Error(types.HttpcError(err))
  }
}

pub fn try_send(
  request: Request(BitArray),
  apply fun: fn(Response(BitArray)) -> ApiResponse(data),
) {
  result.try(send(request), fun)
}
